import pytest
import numpy as np
from titanq._model.numpy_util import (
    convert_to_float32,
    is_upperbound_bigger_equal_lowerbound,
    is_ndarray_binary,
    ndarray_contains_nan_inf,
    reshape_to_2d,
    validate_cardinalities,
)

@pytest.mark.parametrize("array, expected_dtype", [
    (np.array([[1, 2], [3, 4]]), np.float32),
    (np.array([[1.5, 2.5], [3.5, 4.5]]), np.float32),
    (np.empty((0, 0)), np.float32),
    (np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]]), np.float32)
])
def test_convert_to_float32(array: np.ndarray, expected_dtype: np.dtype):
    result = convert_to_float32(array)
    assert result.dtype == expected_dtype


@pytest.mark.parametrize("shape, expected_error", [
    ((4,), None),
    ((1,4), ValueError),
    ((99,), None),
    ((), AttributeError),
    ((99,0), ValueError),
    ((4, 4), ValueError),
    ((0,4), ValueError)
])
def test_reshape_to_2d(shape: np.shape, expected_error):
    reshaped_array = np.random.rand(*shape)

    if expected_error:
        with pytest.raises(expected_error):
            result = reshape_to_2d(reshaped_array)
    else:
        result = reshape_to_2d(reshaped_array)
        assert result.ndim == 2


@pytest.mark.parametrize("array, expected_result", [
    (np.array([2]), False),
    (np.array([2, 0]), False),
    (np.array([-1, 0]), False),
    (np.array([0, 1.01]), False),
    (np.array([[0, 0], [0, 2]]), False),
    (np.array([0]), True),
    (np.array([0, 1]), True),
    (np.array([[1, 1], [0, 0]]), True),
])
def test_is_ndarray_binary(array, expected_result):
    assert is_ndarray_binary(array) == expected_result


@pytest.mark.parametrize("array, expected_value", [
    (np.array([0]), False),
    (np.array([1]), False),
    (np.array([1.01]), False),
    (np.array([-1.01]), False),
    (np.array([np.nan]), True),
    (np.array([-np.nan]), True),
    (np.array([np.NaN]), True),
    (np.array([np.NAN]), True),
    (np.array([np.inf]), True),
    (np.array([-np.inf]), True),
])
def test_ndarray_contains_nan_inf(array, expected_value):
    assert ndarray_contains_nan_inf(array) == expected_value

@pytest.mark.parametrize("array, expected_value", [
    (np.array([[0, 1], [-1, 4000]]), False),
    (np.array([[1, 1], [0, 1]]), True),
    (np.array([[1.01, 1], [5.4, 1]]), True),
])
def test_is_upperbound_bigger_equal_lowerbound(array, expected_value):
    assert is_upperbound_bigger_equal_lowerbound(array) == expected_value


@pytest.mark.parametrize("cardinalities, expected_error", [
    (2, None),
    (3, None),
    (4, ValueError),
])
def test_validate_cardinalities(cardinalities, expected_error):
    if expected_error:
        with pytest.raises(expected_error):
            validate_cardinalities(np.array([[1,1,1]]), np.array([cardinalities]))
    else:
        validate_cardinalities(np.array([[1,1,1]]), np.array([cardinalities]))


import abc
import enum
from typing import List, Tuple
import numpy as np
from numpy._typing import NDArray

class Vtype(str, enum.Enum):
    """
    All variables types currently supported by the solver
    """

    BINARY = 'binary'
    BIPOLAR = 'bipolar'
    INTEGER = 'integer'
    CONTINUOUS = 'continuous'

    def __str__(self) -> str:
        return str(self.value)

class VariableVector(abc.ABC):
    """
    Object That represent a vector of variable to be optimized.
    """
    def __init__(self, name: str, size: int) -> None:
        if size < 1:
            raise ValueError("Variable vector size cannot be less than 1")

        self._name = name
        self._size = size


    def size(self) -> int:
        """
        :return: size of this vector.
        """
        return self._size


    def name(self) -> str:
        """
        :return: Name of this variable vector.
        """
        return self._name


    @abc.abstractmethod
    def vtype(self) -> Vtype:
        """
        :return: Type of variable in the vector.
        """


    @abc.abstractmethod
    def variable_types_as_list(self) -> str:
        """
        :return: Generate a string of 'b', 'i' or 'c' depending on the variable type
        """


    @abc.abstractmethod
    def variable_bounds(self) -> NDArray:
        """
        :return: The variable bounds associated to this variable vector
        """


class BinaryVariableVector(VariableVector):
    def vtype(self) -> Vtype:
        return Vtype.BINARY


    def variable_types_as_list(self) -> str:
        return "b" * self.size()


    def variable_bounds(self) -> NDArray:
        return np.tile(np.array([0,1], dtype=np.float32), (self._size, 1))


# This class violate the Liskov Substitution Principle because it raise error in some of
# the method it need to implement. This is somewhat fine for now but should be fix soon.
class BipolarVariableVector(VariableVector):
    def vtype(self) -> Vtype:
        return Vtype.BIPOLAR

    def variable_types_as_list(self) -> str:
        raise ValueError("Cannot set variable types as a list for 'bipolar' variable type")

    def variable_bounds(self) -> NDArray:
        raise ValueError("Cannot define variable bounds for 'bipolar' variable type")


class IntegerVariableVector(VariableVector):
    def __init__(self, name: str, size: int, variable_bounds: List[Tuple[int, int]]) -> None:
        super().__init__(name, size)

        if len(variable_bounds) != self.size():
            raise ValueError("variable_bounds need to be the same length as variable size")

        self._variable_bounds = np.array(variable_bounds, dtype=np.float32)


    def vtype(self) -> Vtype:
        return Vtype.INTEGER


    def variable_types_as_list(self) -> str:
        return "i" * self.size()


    def variable_bounds(self) -> NDArray:
        return self._variable_bounds

class ContinuousVariableVector(VariableVector):
    def __init__(self, name: str, size: int, variable_bounds: List[Tuple[int, int]]) -> None:
        super().__init__(name, size)

        if len(variable_bounds) != self.size():
            raise ValueError("variable_bounds need to be the same length as variable size")

        self._variable_bounds = np.array(variable_bounds, dtype=np.float32)


    def vtype(self) -> Vtype:
        return Vtype.CONTINUOUS


    def variable_types_as_list(self) -> str:
        return "c" * self.size()


    def variable_bounds(self) -> NDArray:
        return self._variable_bounds


class TitanQError(Exception):
    """Base TitanQ error"""

class MissingVariableError(TitanQError):
    """Variable has not already been registered"""

class VariableAlreadyExist(TitanQError):
    """Variable with the same name already exist"""

class MissingObjectiveError(TitanQError):
    """Objective has not already been registered"""

class MaximumConstraintLimitError(TitanQError):
    """The number of constraints is bigger than the number of variables"""

class ConstraintSizeError(TitanQError):
    """The constraint size does not match"""

class ConstraintAlreadySetError(TitanQError):
    """A constraint has already been set"""

class ObjectiveAlreadySetError(TitanQError):
    """An objective has already been set"""

class OptimizeError(TitanQError):
    """Error occur during optimization"""

class ServerError(TitanQError):
    """Error returned by the server"""

class ConnectionError(TitanQError):
    """Error due to a connection issue with an external resource"""


import logging
import numpy as np
from numpy.typing import NDArray
from typing import Optional

from .errors import ConstraintSizeError, MaximumConstraintLimitError
from .numpy_util import convert_to_float32

log = logging.getLogger("TitanQ")

MAX_CONSTRAINTS_COUNT = 16_000

class Constraints:
    def __init__(self) -> None:
        self._variable_size = 0
        self._constraint_weights = None
        self._constraint_bounds = None


    def is_empty(self) -> bool :
        return self._constraint_bounds is None and self._constraint_weights is None


    def augment_size(self, size: int):
        self._variable_size += size


    def add_constraint(self, constraint_weights: np.ndarray, constraint_bounds: np.ndarray):
        """
        Add a constraint to the existing ones

        :param constraint_weights: constraint_weights to append to the existing ones.
        :param constraint_bounds: constraint_bounds to append to the existing ones.

        :raises ConstraintSizeError: constraint size is different than variable size.
        :raises MaximumConstraintLimitError: the number of constraint exeed the limit.
        """
        if constraint_weights.shape[1] != self._variable_size:
            raise ConstraintSizeError(
                "Constraint mask shape does not match the variable size. " \
                f"Constraint size: {constraint_weights.shape[1]}, Variable size: {self._variable_size}")

        if self._constraints_rows() + constraint_weights.shape[0] > MAX_CONSTRAINTS_COUNT:
            raise MaximumConstraintLimitError(
                "Cannot add additional constraints. The limit of constraints have been reached. " \
                f"Number of constraints: {self._constraints_rows()} while {self._constraints_rows() + constraint_weights.shape[0]} is trying to be added."
            )

        # API only takes np.float32
        constraint_weights = convert_to_float32(constraint_weights)
        constraint_bounds = convert_to_float32(constraint_bounds)

        self._append_constraint_weights(constraint_weights)
        self._append_constraint_bounds(constraint_bounds)


    def weights(self):
        """
        :return: The weights constraints.
        """
        return self._constraint_weights


    def bounds(self):
        """
        :return: The bounds constraints.
        """
        return self._constraint_bounds


    def _append_constraint_weights(self, constraint_weights_to_add: NDArray[np.float32]) -> None:
        """Appends ``constraint_weights_to_add`` to the existing one."""
        if self._constraint_weights is None:
            self._constraint_weights = constraint_weights_to_add
        else:
            self._constraint_weights = np.append(self._constraint_weights, constraint_weights_to_add, axis=0)


    def _append_constraint_bounds(self, constraint_bounds_to_add: NDArray[np.float32]) -> None:
        """Appends ``constraint_bounds_to_add`` to the existing one."""
        if self._constraint_bounds is None:
            self._constraint_bounds = constraint_bounds_to_add
        else:
            self._constraint_bounds =  np.append(self._constraint_bounds, constraint_bounds_to_add, axis=0)


    def _constraints_rows(self) -> int:
        """
        :return: The number of constraints (row) already set, 0 if never set
        """
        if self._constraint_weights is None:
            return 0
        return self._constraint_weights.shape[0]



import datetime
from itertools import chain
from typing import Optional
import boto3
import botocore
import logging
import time

from .._client.model import AwsStorage, S3Input, S3Output
from .storage_client import StorageClient

log = logging.getLogger("TitanQ")

_BIAS_FILE = "bias.npy"
_WEIGHTS_FILE = "weights.npy"
_CONSTRAINT_BOUNDS_FILE = "constraint_bounds.npy"
_CONSTRAINT_WEIGHTS_FILE = "constraints_weights.npy"
_VARIABLE_BOUNDS_FILE = "variable_bounds.npy"
_RESULT_FILE = "result.zip"



class S3Storage(StorageClient):
    """Storage client using S3 bucket from AWS"""
    def __init__(
        self,
        access_key: str,
        secret_key: str,
        bucket_name: str,
    ) -> None:
        """
        Initiate the S3 bucket client for handling temporary files.

        Parameters
        ----------
        access_key
            Used to upload and download files from an AWS S3 bucket.
        secret_key
            Used to upload and download files from an AWS S3 bucket.
        bucket_name
            Name of the AWS S3 bucket used to store temporarily data that the TitanQ optimizer will read.

        Raises
        ------
        botocore.exceptions.ParamValidationError
            If any AWS argument is missing this will raise an exception.

        Examples
        --------
        >>> storage_client = S3Storage(
        >>>     access_key="{insert aws bucket access key here}",
        >>>     secret_key="{insert aws bucket secret key here}",
        >>>     bucket_name="{insert bucket name here}"
        >>> )
        """
        self._s3 = boto3.client('s3', aws_access_key_id=access_key, aws_secret_access_key=secret_key)
        self._access_key_id = access_key
        self._secret_access_key = secret_key
        self._bucket_name = bucket_name

        timestamp = datetime.datetime.now().isoformat()
        self._remote_folder = f"titanq_sdk/{timestamp}"

        # keep track of which file were uploaded
        self._file_uploaded = set()

    def _upload_arrays(
        self,
        bias: bytes,
        weights: bytes,
        constraint_bounds: Optional[bytes],
        constraint_weights: Optional[bytes],
        variable_bounds: Optional[bytes]
    ) -> None:

        upload_tuple = [
            (self._get_full_filename(_BIAS_FILE), bias),
            (self._get_full_filename(_WEIGHTS_FILE), weights)
        ]

        if constraint_bounds:
            upload_tuple.append((self._get_full_filename(_CONSTRAINT_BOUNDS_FILE), constraint_bounds))

        if constraint_weights:
            upload_tuple.append((self._get_full_filename(_CONSTRAINT_WEIGHTS_FILE), constraint_weights))

        if variable_bounds:
            upload_tuple.append((self._get_full_filename(_VARIABLE_BOUNDS_FILE), variable_bounds))

        for filename, body in upload_tuple:
            log.debug(f"Uploading object on AWS s3: {filename}")
            self._s3.put_object(Body=body, Bucket=self._bucket_name, Key=filename)
            self._file_uploaded.add(filename)

    def _input(self) -> S3Input:
        constraint_bounds = self._get_full_filename(_CONSTRAINT_BOUNDS_FILE)
        constraint_weights = self._get_full_filename(_CONSTRAINT_WEIGHTS_FILE)
        variable_bounds = self._get_full_filename(_VARIABLE_BOUNDS_FILE)

        return S3Input(
            s3=self._get_api_model_location(),
            weights_file_name=self._get_full_filename(_WEIGHTS_FILE),
            bias_file_name=self._get_full_filename(_BIAS_FILE),
            constraint_weights_file_name=constraint_weights if constraint_weights in self._file_uploaded else None,
            constraint_bounds_file_name=constraint_bounds if constraint_bounds in self._file_uploaded else None,
            variable_bounds_file_name=variable_bounds if variable_bounds in self._file_uploaded else None,
            manifest=None
        )

    def _output(self) -> S3Output:
        return S3Output(
            result_archive_file_name=self._get_full_filename(_RESULT_FILE),
            s3=self._get_api_model_location())

    def _get_api_model_location(self) -> AwsStorage:
        """
        :return: An AwsStorage object that can be used in the api_model using this S3 credentials
        """
        return AwsStorage(
            bucket_name=self._bucket_name,
            access_key_id=self._access_key_id,
            secret_access_key=self._secret_access_key,
        )

    def _wait_for_result_to_be_uploaded_and_download(self) -> bytes:
        result_file = self._get_full_filename(_RESULT_FILE)
        self._wait_for_file_to_be_uploaded(result_file)
        return self._download_file(result_file)

    def _wait_for_file_to_be_uploaded(self, filename: str):
        """
        Wait until a file exist in a bucket. It also verifies if the
        file is bigger than 0 bytes, this will ensure not downloading
        the empty archive file uploaded to test credentials

        :param filename: The full path of the file that is uploaded
        """
        log.debug(f"Waiting until object get upload on AWS s3: {filename}")
        while True:
            try:
                # check if file exist in the s3 bucket
                response = self._s3.head_object(
                    Bucket=self._bucket_name,
                    Key=filename,
                )
                # check if file content_length > 0
                if response['ContentLength'] > 0:
                    break
            except botocore.exceptions.ClientError as ex:
                # if the error we got is not 404, This is an unexpected error. Raise it
                if int(ex.response['Error']['Code']) != 404:
                    raise

            time.sleep(0.25) # wait 0.25 sec before trying again

    def _download_file(self, filename) -> bytes:
        """
        Download file from remote s3 bucket

        :param filename: The full path of the file to be uploaded

        :return: content of the file
        """
        log.debug(f"Downloading object from AWS s3: {filename}")
        object = self._s3.get_object(Bucket=self._bucket_name, Key=filename)
        return object['Body'].read()

    def _delete_remote_object(self):
        """
        Delete remote object on AWS s3

        :param key: object name to be deleted on the remote s3 bucket
        """
        for file_name in chain(self._file_uploaded, [self._get_full_filename(_RESULT_FILE)]):
            log.debug(f"Deleting object on AWS s3: {file_name}")
            self._s3.delete_object(Bucket=self._bucket_name, Key=file_name)

    def _get_full_filename(self, filename: str) -> str:
        return f"{self._remote_folder}/{filename}"


import numpy as np

from titanq import OptimizeResponse
from titanq._model.variable_list import VariableVectorList
from titanq._model.variable import BinaryVariableVector

def test_multiple_variables():
    result = np.array([
        [0, 1, 1, 0, 1, 0],
        [1, 0, 0, 0, 0, 1],
    ], dtype=np.float32)
    variable_list = VariableVectorList()
    variable_list.add(BinaryVariableVector('x', 4))
    variable_list.add(BinaryVariableVector('y', 2))

    response = OptimizeResponse(variable_list, result, {})

    expected_x = np.array([
        [0, 1, 1, 0],
        [1, 0, 0, 0],
    ], dtype=np.float32)

    expected_y = np.array([
        [1, 0],
        [0, 1],
    ], dtype=np.float32)

    assert np.array_equal(response.x, expected_x)
    assert np.array_equal(response.y, expected_y)
    assert np.array_equal(response.result_vector(), result)


def test_get_metrics():
    metrics = { "computation_metrics": { "metrics1": 1, "metrics2": "value2" }}
    variable_list = VariableVectorList()
    variable_list.add(BinaryVariableVector('x', 6))
    response = OptimizeResponse(variable_list,  np.random.rand(2, 6).astype(np.float32), metrics)

    print(response.computation_metrics("metrics1"))

    assert response.computation_metrics("metrics1") == metrics["computation_metrics"]["metrics1"]
    assert response.computation_metrics("metrics2") == metrics["computation_metrics"]["metrics2"]



import logging
import numpy as np
from numpy.typing import NDArray

log = logging.getLogger("TitanQ")


def convert_to_float32(array: np.ndarray) -> NDArray[np.float32]:
    """
    Changes a given array type to np.NDArray[np.float32].

    :param array: input constraint mask of shape (M, N).
    :type array: a NumPy 2-D dense ndarray.

    :return: Array of type float32
    :rtype: NDArray[np.float32]
    """
    return array.astype(np.float32)


def reshape_to_2d(array: np.ndarray) -> np.ndarray:
    """
    Reshape a given array to a 2-D array.

    :param array: input constraint mask of shape (N,).
    :type array: a NumPy 1-D dense ndarray.

    :return: Array of type Any
    :rtype: NDArray[np.Any]
    """
    if array.ndim != 1:
        raise ValueError("Input array is not 1-D. Cannot reshape array.")
    return array.reshape(1, -1)


def is_ndarray_binary(array: np.ndarray) -> bool:
    """
    :return: if all values ares either 1 or 0 (binary)
    """
    return np.all((array == 1) | (array == 0))


def ndarray_contains_nan_inf(array: np.ndarray) -> bool:
    """
    :return: if the array contains NaN or inf type
    """
    return np.isnan(array).any() or np.isinf(array).any()

def is_upperbound_bigger_equal_lowerbound(array: np.ndarray) -> bool:
    """
    finds lowerbounds larger or equal than upperbounds in NumPy 2-D Array (M, 2)
    where M is the number of rows

    :return: if each value of the left side is lower lower than the right side
    """
    left_slice = array[:, 0]
    right_slice = array[:, 1]

    equal_or_higher = np.where(left_slice >= right_slice)[0]
    return len(equal_or_higher) > 0

def validate_cardinalities(constraint_mask: np.ndarray, cardinalities: np.ndarray) -> None:
    """
    Validates each row's sum of a binary array against corresponding values in a cardinalities array.
    It prints warnings if the sum is equal to the cardinality and raises an error if the sum is less than the cardinality.

    :param array: input constraint mask of shape (M, N).
    :type array: a NumPy 2-D dense ndarray.

    :param cardinalities: The constraint_rhs vector of shape (M,) where M is the number of constraints.
    :type cardinalities:  a NumPy 1-D ndarray.
    """
    binary_sums = np.sum(constraint_mask, axis=1)

    equal_indices = np.where(binary_sums == cardinalities[0])[0]
    less_indices = np.where(binary_sums < cardinalities)[0]

    if equal_indices.size > 0:
        log.warning(f" The sum of rows {', '.join(map(str, equal_indices))} in the binary array equals its corresponding cardinality.")

    if less_indices.size > 0:
        raise ValueError(f"The sum of rows {', '.join(map(str, less_indices))} in the binary array is less than its corresponding cardinality.")


import datetime
import io
import json
import uuid
import numpy as np
import pytest
from typing import Any, Dict, List
import zipfile

from titanq import Model, Vtype, errors, Target
from titanq._storage.s3_storage import S3Storage
from titanq._storage.managed_storage import ManagedStorage
from titanq._client.model import SolveResponse

from .mock import TitanQClientMock, MockS3StorageClient


def file_in_filename_list(filename: str, filename_list: List[str]) -> bool:
    return any(filename in s for s in filename_list)

def np_array_to_bytes(arr: np.ndarray) -> bytes:
    buf = io.BytesIO()
    np.save(buf, arr)
    return buf.getvalue()

@pytest.fixture
def mock_metrics() -> Dict[str, Any]:
    return {
        "computation_metrics": {
            "metrics1": 1,
            "metrics2": "value2"
        }
    }

@pytest.fixture
def mock_result() -> Dict[str, Any]:
    return [np.random.rand(10).astype(np.float32) for _ in range(10)]

@pytest.fixture
def mock_s3_storage(mock_metrics, mock_result) -> MockS3StorageClient:
    # expected npy result file content
    expected_result_buffer = io.BytesIO()
    np.save(expected_result_buffer, mock_result)

    # set mock client with mock values
    buff = io.BytesIO()
    with zipfile.ZipFile(buff, 'w') as file:
        file.writestr("result.npy", expected_result_buffer.getvalue())
        file.writestr("metrics.json", json.dumps(mock_metrics).encode())

    return MockS3StorageClient(buff.getvalue())

@pytest.fixture
def mock_titanq_client() -> TitanQClientMock:
    return TitanQClientMock(solve_response=SolveResponse(
        computation_id=str(uuid.uuid4()),
        status="queued",
        message="Computation have been queued"
    ))

@pytest.fixture
def model_s3_client(mock_s3_storage, mock_titanq_client) -> Model:
    model = Model(
        api_key="test_api_key",
        storage_client=mock_s3_storage
    )

    model._titanq_client = mock_titanq_client
    return model


@pytest.fixture
def constant_datetime(monkeypatch):
    constant_datetime = datetime.datetime(2024,1,1,8,0,0)
    class MockDatetime(datetime.datetime):
        @classmethod
        def now(cls):
            return constant_datetime

    monkeypatch.setattr(datetime, 'datetime', MockDatetime)
    return constant_datetime

@pytest.mark.parametrize("api_key, storage_client ,expected_storage_class", [
    ("api_key", S3Storage(access_key="aws_access_key", secret_key="aws_secret_access_key", bucket_name="bucket_name"), S3Storage),
    ("api_key", ManagedStorage(TitanQClientMock()), ManagedStorage),
])
def test_selected_storage(api_key, storage_client, expected_storage_class):
    model = Model(api_key=api_key, storage_client=storage_client)
    assert isinstance(model._storage_client, expected_storage_class)


@pytest.mark.parametrize("name, size, vtype, error", [
    ('x', 1, Vtype.BINARY, None),
    ('x', 47, Vtype.BINARY, None),
    ('x', -1, Vtype.BINARY, ValueError),
    ('x', 0, Vtype.BINARY, ValueError)
])
def test_new_variable(model_s3_client, name, size, vtype, error):
    if error:
        with pytest.raises(error):
            model_s3_client.add_variable_vector(name, size, vtype)
    else:
        model_s3_client.add_variable_vector(name, size, vtype)


@pytest.mark.parametrize("weights_shape, bias_shape, objective, error", [
    ((10, 10), (10,), Target.MINIMIZE, None),
    ((11, 10), (10,), Target.MINIMIZE, ValueError),
    ((10, 11), (10,), Target.MINIMIZE, ValueError),
    ((11, 11), (10,), Target.MINIMIZE, ValueError),
    ((10, 10, 10), (10,), Target.MINIMIZE, ValueError),
    ((10,), (10,), Target.MINIMIZE, ValueError),
    ((10,10), (9,), Target.MINIMIZE, ValueError),
    ((10,10), (10,1), Target.MINIMIZE, ValueError),
    ((10,10), (10,2), Target.MINIMIZE, ValueError),
])
def test_set_objective(model_s3_client, weights_shape, bias_shape, objective, error):
    model_s3_client.add_variable_vector('x', 10, Vtype.BINARY)

    weights = np.random.rand(*weights_shape).astype(np.float32)
    bias = np.random.rand(*bias_shape).astype(np.float32)

    if error:
        with pytest.raises(error):
            model_s3_client.set_objective_matrices(weights, bias, objective)
    else:
        model_s3_client.set_objective_matrices(weights, bias, objective)

@pytest.mark.parametrize("weights_data_type, bias_data_type, error", [
    (np.float32, np.float32, None),
    (np.float64, np.float32, ValueError),
    (np.float32, np.float64, ValueError),
    (np.int32, np.float32, ValueError),
    (np.float32, np.int32, ValueError),
    (np.bool_, np.float32, ValueError),
    (np.float32, np.bool_, ValueError),
    (np.byte, np.float32, ValueError),
    (np.float32, np.byte, ValueError),
    (np.short, np.float32, ValueError),
    (np.float32, np.short, ValueError),
])
def test_objective_matrices_data_type(model_s3_client, weights_data_type, bias_data_type, error):
    model_s3_client.add_variable_vector('x', 10, Vtype.BINARY)

    weights = np.random.rand(10, 10).astype(weights_data_type)
    bias = np.random.rand(10).astype(bias_data_type)

    if error:
        with pytest.raises(error):
            model_s3_client.set_objective_matrices(weights, bias, Target.MINIMIZE)
    else:
        model_s3_client.set_objective_matrices(weights, bias, Target.MINIMIZE)


def test_set_objective_without_variable(model_s3_client):
    weights = np.random.rand(10, 10)
    bias = np.random.rand(10)

    with pytest.raises(errors.MissingVariableError):
        model_s3_client.set_objective_matrices(weights, bias, Target.MINIMIZE)


def test_set_2_objective(model_s3_client):
    model_s3_client.add_variable_vector('x', 10, Vtype.BINARY)

    weights = np.random.rand(10, 10).astype(np.float32)
    bias = np.random.rand(10).astype(np.float32)

    model_s3_client.set_objective_matrices(weights, bias, Target.MINIMIZE)

    with pytest.raises(errors.ObjectiveAlreadySetError):
        model_s3_client.set_objective_matrices(weights, bias, Target.MINIMIZE)


def test_optimize_no_variable(model_s3_client):
    with pytest.raises(errors.MissingVariableError):
        model_s3_client.optimize()


def test_optimize_no_objective(model_s3_client):
    model_s3_client.add_variable_vector('x', 10, Vtype.BINARY)

    with pytest.raises(errors.MissingObjectiveError):
        model_s3_client.optimize()


def test_optimize_no_constraints(model_s3_client, mock_s3_storage, mock_metrics, mock_result):
    model = model_s3_client

    # optimize using sdk
    weights = np.random.rand(10, 10).astype(np.float32)
    bias = np.random.rand(10).astype(np.float32)

    model.add_variable_vector('x', 10, Vtype.BINARY)
    model.set_objective_matrices(weights, bias)
    response = model.optimize()

    assert response.computation_metrics() == mock_metrics["computation_metrics"]
    assert (response.result_vector() == mock_result).all()

    assert mock_s3_storage.array_uploaded['bias'] == np_array_to_bytes(bias)
    assert mock_s3_storage.array_uploaded['weights'] == np_array_to_bytes(weights)
    assert mock_s3_storage.array_uploaded['constraint_weights'] == None
    assert mock_s3_storage.array_uploaded['constraint_bounds'] == None

    assert mock_s3_storage.file_removed


@pytest.mark.parametrize("vtype", [
    (Vtype.BINARY),
    (Vtype.BIPOLAR),
])
def test_vtype_sent(model_s3_client, mock_titanq_client, vtype):
    model = model_s3_client
    mock_client = mock_titanq_client

    # optimize using sdk
    weights = np.random.rand(10, 10).astype(np.float32)
    bias = np.random.rand(10).astype(np.float32)

    model.add_variable_vector('x', 10, vtype)
    model.set_objective_matrices(weights, bias)

    model.optimize()

    assert mock_client.request_sent.parameters.variable_types == str(vtype)


def test_constraints_without_variable(model_s3_client):
    constraint_weights = np.random.rand(2, 10).astype(np.float32)
    constraint_bounds = np.random.rand(2).astype(np.float32)

    with pytest.raises(errors.MissingVariableError):
        model_s3_client.add_set_partitioning_constraints_matrix(np.array([[0, 0, 1, 0], [0, 1, 0, 0]]))

    with pytest.raises(errors.MissingVariableError):
        model_s3_client.add_set_partitioning_constraint(np.array([0, 1]))

    with pytest.raises(errors.MissingVariableError):
        model_s3_client.add_cardinality_constraints_matrix(np.array([[1, 0, 1, 0], [0, 1, 1, 1]]), np.array([2, 3]))

    with pytest.raises(errors.MissingVariableError):
        model_s3_client.add_cardinality_constraint(np.array([1, 1]), 2)


def test_add_set_partitioning_constraints_matrix(model_s3_client):
    model_s3_client.add_variable_vector('x', 4, Vtype.BINARY)

    model_s3_client.add_set_partitioning_constraints_matrix(np.array([[0, 0, 1, 0], [0, 1, 0, 0]]))
    np.testing.assert_equal(model_s3_client._constraints.weights(), np.array([[0, 0, 1, 0], [0, 1, 0, 0]]))
    np.testing.assert_equal(model_s3_client._constraints.bounds(), np.array([[1, 1], [1, 1]]))

    model_s3_client.add_set_partitioning_constraints_matrix(np.array([[0, 1, 0, 0], [1, 0, 0, 0]]))
    np.testing.assert_equal(model_s3_client._constraints.weights(), np.array([[0, 0, 1, 0], [0, 1, 0, 0], [0, 1, 0, 0], [1, 0, 0, 0]]))
    np.testing.assert_equal(model_s3_client._constraints.bounds(), np.array([[1, 1], [1, 1], [1, 1], [1, 1]]))


def test_add_set_partitioning_constraint(model_s3_client):
    model_s3_client.add_variable_vector('x', 2, Vtype.BINARY)

    model_s3_client.add_set_partitioning_constraint(np.array([0, 1]))
    np.testing.assert_equal(model_s3_client._constraints.weights(), np.array([[0, 1]]))
    np.testing.assert_equal(model_s3_client._constraints.bounds(), np.array([[1, 1]]))

    model_s3_client.add_set_partitioning_constraint(np.array([1, 0]))
    np.testing.assert_equal(model_s3_client._constraints.weights(), np.array([[0, 1], [1, 0]]))
    np.testing.assert_equal(model_s3_client._constraints.bounds(), np.array([[1, 1], [1, 1]]))


def test_add_cardinality_constraints_matrix(model_s3_client):
    model_s3_client.add_variable_vector('x', 4, Vtype.BINARY)

    model_s3_client.add_cardinality_constraints_matrix(np.array([[1, 0, 1, 0], [0, 1, 1, 1]]), np.array([2, 3]))
    np.testing.assert_equal(model_s3_client._constraints.weights(), np.array([[1, 0, 1, 0], [0, 1, 1, 1]]))
    np.testing.assert_equal(model_s3_client._constraints.bounds(), np.array([[2, 2], [3, 3]]))

    model_s3_client.add_cardinality_constraints_matrix(np.array([[0, 1, 0, 0], [1, 0, 0, 0]]), np.array([1, 1]))
    np.testing.assert_equal(model_s3_client._constraints.weights(), np.array([[1, 0, 1, 0], [0, 1, 1, 1], [0, 1, 0, 0], [1, 0, 0, 0]]))
    np.testing.assert_equal(model_s3_client._constraints.bounds(), np.array([[2, 2], [3, 3], [1, 1], [1, 1]]))

def test_add_cardinality_constraint(model_s3_client):
    model_s3_client.add_variable_vector('x', 2, Vtype.BINARY)

    model_s3_client.add_cardinality_constraint(np.array([1, 1]), 2)
    np.testing.assert_equal(model_s3_client._constraints.weights(), np.array([[1, 1]]))
    np.testing.assert_equal(model_s3_client._constraints.bounds(), np.array([[2, 2]]))

    model_s3_client.add_cardinality_constraint(np.array([0, 1]), 1)
    np.testing.assert_equal(model_s3_client._constraints.weights(), np.array([[1, 1], [0, 1]]))
    np.testing.assert_equal(model_s3_client._constraints.bounds(), np.array([[2, 2], [1, 1]]))

@pytest.mark.parametrize("constraint_mask, cardinality, error", [
    (np.array([1, 1]), 1, None),
    (np.array([1, 1]), 2, None),
    (np.array([1, 1]), 3, ValueError),
])
def test_cardinalities_sum(model_s3_client, constraint_mask, cardinality, error):
    model_s3_client.add_variable_vector('x', 2, Vtype.BINARY)

    if error:
        with pytest.raises(ValueError):
            model_s3_client.add_cardinality_constraint(constraint_mask, cardinality)
    else:
        model_s3_client.add_cardinality_constraint(constraint_mask, cardinality)

def test_cardinality_or_set_partitioning_constraint_variable_types(model_s3_client, mock_titanq_client):
    weights = np.random.rand(2, 2).astype(np.float32)
    bias = np.random.rand(2).astype(np.float32)

    model_s3_client.add_variable_vector('x', 2, Vtype.BINARY)
    model_s3_client.set_objective_matrices(weights, bias)
    model_s3_client.add_cardinality_constraint(np.array([1, 1]), 2)

    model_s3_client.optimize()

    assert mock_titanq_client.request_sent.parameters.variable_types == "binary"
    assert mock_titanq_client.request_sent.input.variable_bounds_file_name == 'test-variable_bounds.npy'

def test_add_equality_constraint(model_s3_client):
    model_s3_client.add_variable_vector('x', 2, Vtype.BINARY)

    model_s3_client.add_equality_constraint(np.array([1.05, -1.1], dtype=np.float32), -3.45)
    np.testing.assert_equal(model_s3_client._constraints.weights(), np.array([[1.05, -1.1]], dtype=np.float32))
    np.testing.assert_equal(model_s3_client._constraints.bounds(), np.array([[-3.45, -3.45]], dtype=np.float32))

    model_s3_client.add_equality_constraint(np.array([14, 0], dtype=np.float32), 0)
    np.testing.assert_equal(model_s3_client._constraints.weights(), np.array([[1.05, -1.1], [14, 0]], dtype=np.float32))
    np.testing.assert_equal(model_s3_client._constraints.bounds(), np.array([[-3.45, -3.45], [0, 0]], dtype=np.float32))


@pytest.mark.parametrize("method, kwargs", [
    ("add_set_partitioning_constraints_matrix", { "constraint_mask": np.array([[0,1]])}),
    ("add_set_partitioning_constraint",         { "constraint_mask": np.array([0,1])}),
    ("add_cardinality_constraints_matrix",      { "constraint_mask": np.array([[0,1]]),                   "cardinalities": np.array([1])}),
    ("add_cardinality_constraint",              { "constraint_mask": np.array([0,1]),                     "cardinality": 1 }),
    ("add_equality_constraints_matrix",         { "constraint_mask": np.array([[0,1]], dtype=np.float32), "limit": np.array([1], dtype=np.float32) }),
    ("add_equality_constraint",                 { "constraint_mask": np.array([0,1], dtype=np.float32),   "limit": 1 }),
    ("add_inequality_constraints_matrix",       { "constraint_mask": np.array([[0,1]], dtype=np.float32), "constraint_bounds": np.array([[-50, np.nan]], dtype=np.float32)}),
    ("add_inequality_constraint",               { "constraint_mask": np.array([0,1], dtype=np.float32),   "constraint_bounds": np.array([-50, np.nan], dtype=np.float32)}),
])
def test_constraint_weights_and_bounds_float_32_output(model_s3_client, method, kwargs):
    model_s3_client.add_variable_vector('x', 2, Vtype.BINARY)

    constraint_function = getattr(model_s3_client, method)
    constraint_function(**kwargs)

    assert model_s3_client._constraints.bounds().dtype == np.float32


def test_add_equality_constraint_matrix(model_s3_client):
    model_s3_client.add_variable_vector('x', 4, Vtype.BINARY)

    model_s3_client.add_equality_constraints_matrix(np.array([[-3.51, 0, 0, 0], [10, 0, 0, 0]], dtype=np.float32), np.array([2, 10], dtype=np.float32))
    np.testing.assert_equal(model_s3_client._constraints.weights(), np.array([[-3.51, 0, 0, 0], [10, 0, 0, 0]], dtype=np.float32))
    np.testing.assert_equal(model_s3_client._constraints.bounds(), np.array([[2, 2], [10, 10]], dtype=np.float32))

    model_s3_client.add_equality_constraints_matrix(np.array([[0, 5, 0, 0], [0, 0, -5, 0]], dtype=np.float32), np.array([-1, 1], dtype=np.float32))
    np.testing.assert_equal(model_s3_client._constraints.weights(), np.array([[-3.51, 0, 0, 0], [10, 0, 0, 0], [0, 5, 0, 0], [0, 0, -5, 0]], dtype=np.float32))
    np.testing.assert_equal(model_s3_client._constraints.bounds(), np.array([[2, 2], [10, 10], [-1, -1], [1, 1]], dtype=np.float32))

    with pytest.raises(ValueError):
        model_s3_client.add_equality_constraints_matrix(np.array([[1, 2, 3, 4]], dtype=np.float32), np.array([[10, 10]], dtype=np.float32))

    with pytest.raises(ValueError):
        model_s3_client.add_equality_constraints_matrix(np.array([[1, 2, 3, 4]], dtype=np.float32), np.array([10], dtype=np.int32))


def test_equality_constraint_override_variable_types(model_s3_client, mock_titanq_client):
    weights = np.random.rand(2, 2).astype(np.float32)
    bias = np.random.rand(2).astype(np.float32)

    model_s3_client.add_variable_vector('x', 2, Vtype.BINARY)
    model_s3_client.set_objective_matrices(weights, bias)
    model_s3_client.add_equality_constraint(np.array([1.05, -1.1], dtype=np.float32), -3.45)
    model_s3_client.optimize()

    assert mock_titanq_client.request_sent.parameters.variable_types == "bb"
    assert mock_titanq_client.request_sent.input.variable_bounds_file_name == "test-variable_bounds.npy"


def test_add_inequality_constraint(model_s3_client):
    model_s3_client.add_variable_vector('x', 2, Vtype.BINARY)

    model_s3_client.add_inequality_constraint(np.array([1.05, -1.1], dtype=np.float32), np.array([np.nan, -3.45], dtype=np.float32))
    np.testing.assert_equal(model_s3_client._constraints.weights(), np.array([[1.05, -1.1]], dtype=np.float32))
    np.testing.assert_equal(model_s3_client._constraints.bounds(), np.array([[np.nan, -3.45]], dtype=np.float32))

    model_s3_client.add_inequality_constraint(np.array([14, 0], dtype=np.float32), np.array([-0.09, 0], dtype=np.float32))
    np.testing.assert_equal(model_s3_client._constraints.weights(), np.array([[1.05, -1.1], [14, 0]], dtype=np.float32))
    np.testing.assert_equal(model_s3_client._constraints.bounds(), np.array([[np.nan, -3.45], [-0.09, 0]], dtype=np.float32))

    model_s3_client.add_inequality_constraint(np.array([76, -4.8], dtype=np.float32), np.array([None, -8], dtype=np.float32))
    np.testing.assert_equal(model_s3_client._constraints.weights(), np.array([[1.05, -1.1], [14, 0], [76, -4.8]], dtype=np.float32))
    np.testing.assert_equal(model_s3_client._constraints.bounds(), np.array([[np.nan, -3.45], [-0.09, 0], [None, -8]], dtype=np.float32))

def test_add_inequality_constraint_matrix(model_s3_client):
    model_s3_client.add_variable_vector('x', 2, Vtype.BINARY)
    model_s3_client.add_inequality_constraints_matrix(np.array([[-3.51, 0], [10, 0]], dtype=np.float32), np.array([[8, 9], [np.nan, 100_000]], dtype=np.float32))

    np.testing.assert_equal(model_s3_client._constraints.weights(), np.array([[-3.51, 0], [10, 0]], dtype=np.float32))
    np.testing.assert_equal(model_s3_client._constraints.bounds(), np.array([[8, 9], [np.nan, 100_000]], dtype=np.float32))

    with pytest.raises(ValueError):
        model_s3_client.add_equality_constraints_matrix(np.array([[1, 2]], dtype=np.float32), np.array([[np.nan, 10]], dtype=np.float32))

    with pytest.raises(ValueError):
        model_s3_client.add_equality_constraints_matrix(np.array([[1, 2]], dtype=np.float32), np.array([np.nan, 10], dtype=np.int32))

def test_inequality_constraint_override_variable_types(model_s3_client, mock_titanq_client):
    weights = np.random.rand(2, 2).astype(np.float32)
    bias = np.random.rand(2).astype(np.float32)

    model_s3_client.add_variable_vector('x', 2, Vtype.BINARY)
    model_s3_client.set_objective_matrices(weights, bias)
    model_s3_client.add_inequality_constraint(np.array([1.05, -1.1], dtype=np.float32), np.array([1.0, np.nan], dtype=np.float32))
    model_s3_client.optimize()

    assert mock_titanq_client.request_sent.parameters.variable_types == "bb"
    assert mock_titanq_client.request_sent.input.variable_bounds_file_name == "test-variable_bounds.npy"


def test_combination_constraint(model_s3_client):
    model_s3_client.add_variable_vector('x', 4, Vtype.BINARY)

    model_s3_client.add_set_partitioning_constraints_matrix(np.array([[0, 0, 1, 0], [0, 1, 0, 0]]))
    np.testing.assert_equal(model_s3_client._constraints.weights(), np.array([[0, 0, 1, 0], [0, 1, 0, 0]]))
    np.testing.assert_equal(model_s3_client._constraints.bounds(), np.array([[1, 1], [1, 1]]))

    model_s3_client.add_set_partitioning_constraint(np.array([0, 1, 0, 0]))
    np.testing.assert_equal(model_s3_client._constraints.weights(), np.array([[0, 0, 1, 0], [0, 1, 0, 0], [0, 1, 0, 0]]))
    np.testing.assert_equal(model_s3_client._constraints.bounds(), np.array([[1, 1], [1, 1], [1, 1]]))

    model_s3_client.add_cardinality_constraint(np.array([1, 1, 1, 1]), 2)
    np.testing.assert_equal(model_s3_client._constraints.weights(), np.array([[0, 0, 1, 0], [0, 1, 0, 0], [0, 1, 0, 0], [1, 1, 1, 1]]))
    np.testing.assert_equal(model_s3_client._constraints.bounds(), np.array([[1, 1], [1, 1], [1, 1], [2, 2]]))


@pytest.mark.parametrize("nmb_of_constraints, expected_error", [
    (15_999, None),
    (16_000, None),
    (16_001, errors.MaximumConstraintLimitError),
])
def test_number_of_constraints_threshold(model_s3_client, nmb_of_constraints, expected_error):
    i = 0
    model_s3_client.add_variable_vector('y', 2, Vtype.BINARY)

    if expected_error:
        with pytest.raises(expected_error):
            while (i < nmb_of_constraints):
                model_s3_client.add_set_partitioning_constraint(np.array([0, 1]))
                i += 1
    else:
        while (i < nmb_of_constraints):
            model_s3_client.add_set_partitioning_constraint(np.array([0, 1]))
            i += 1


def test_manifest(model_s3_client, mock_titanq_client):
    weights = np.random.rand(4, 4).astype(np.float32)
    bias = np.random.rand(4).astype(np.float32)

    # Construct the request
    model_s3_client.add_variable_vector('x', 4, Vtype.BINARY)
    model_s3_client.set_objective_matrices(weights, bias)

    model_s3_client.add_set_partitioning_constraint(np.array([0, 0, 1, 0]))
    model_s3_client.add_cardinality_constraint(np.array([1, 0, 1, 0]), 2)
    model_s3_client.add_equality_constraint(np.array([44.01, 0, 10, -1], dtype=np.float32), -50.01)
    model_s3_client.add_inequality_constraint(np.array([44.01, 0, 10, -1], dtype=np.float32), np.array([np.nan, 10], dtype=np.float32))

    # Send the request
    model_s3_client.optimize()

    # Make sure the appropriate manifest flag have been sent
    assert mock_titanq_client.request_sent.input.manifest.has_set_partitioning_constraint == True
    assert mock_titanq_client.request_sent.input.manifest.has_cardinality_constraint == True
    assert mock_titanq_client.request_sent.input.manifest.has_equality_constraint == True
    assert mock_titanq_client.request_sent.input.manifest.has_inequality_constraint == True


# on titanq package hover when imported, this will show as tooltip in most modern IDE (this comment won't)

"""
TitanQ
======

The TitanQ SDK for python. This package will let you use InfinityQ's QUBO solver named TitanQ


Documentation
-------------

* Documentation is avaiable as sphinx docstrings
* A Quickstart example will be soon available


License
-----------------------------

Apache Software License (Apache 2.0)

"""

# These symbols must be exposed by this lib
from ._model import Model, Vtype, errors, Target, OptimizeResponse

# S3Storage must be exposed to the end user if he wishes to use s3 buckets
from ._storage import S3Storage

# logger config
import logging as _logging
_logging.getLogger("TitanQ").addHandler(_logging.NullHandler())

__version__ = 'v0.9.5'


import io
import json
import logging
import numpy as np
import zipfile
from typing import Any, Dict, List, Tuple, Union

from .constraints import Constraints
from .errors import (
    ConstraintAlreadySetError,
    MissingVariableError,
    MissingObjectiveError,
    ObjectiveAlreadySetError,
    OptimizeError
)
from .objective import Objective, Target
from .optimize_response import OptimizeResponse
from .manifest import Manifest
from .variable import BinaryVariableVector, BipolarVariableVector, ContinuousVariableVector, IntegerVariableVector, Vtype
from .._client import api_model, Client
from .._storage import ManagedStorage, StorageClient
from .numpy_util import (
    is_ndarray_binary,
    is_upperbound_bigger_equal_lowerbound,
    ndarray_contains_nan_inf,
    reshape_to_2d,
    validate_cardinalities
)
from .variable_list import VariableVectorList

log = logging.getLogger("TitanQ")


class Model:
    """
    Root object to define a problem to be optimized
    """

    def __init__(
        self,
        *,
        api_key: str,
        storage_client: StorageClient = None,
        base_server_url: str = "https://titanq.infinityq.io"
        ) -> None:
        """
        Initiate the model with a storage client. If the storage_client is missing, the storage will be managed by TitanQ.

        Notes
        -----
        The storage manged by TitanQ supports weight matrices with a size up to 10k only.

        Parameters
        ----------
        api_key
            TitanQ API key to access the service.
        storage_client
            Storage to choose in order to store some items.
        base_server_url
            TitanQ API server url, default set to ``https://titanq.infinityq.io``.

        Examples
        --------
        With an S3 storage client
            >>> from titanq import Model, S3Storage
            >>> api_key = "{insert API key here}",
            >>> storage_client = S3Storage(
                access_key="{insert aws bucket access key here}",
                secret_key="{insert aws bucket secret key here}",
                bucket_name="{insert bucket name here}"
            )
            >>> model = Model(api_key, storage_client)

        Managed storage client
            >>> from titanq import Model, S3Storage
            >>> api_key = "{insert API key here}",
            >>> model = Model(api_key)
        """
        self._variables = VariableVectorList()
        self._objective: Objective = None
        self._constraints = Constraints()
        self._titanq_client = Client(api_key, base_server_url)
        self._manifest: Manifest = Manifest()

        # the user chose a managed storage or left it as default
        if storage_client is None:
            storage_client = ManagedStorage(self._titanq_client)

        self._storage_client = storage_client


    def add_variable_vector(
        self,
        name: str = '',
        size: int = 1,
        vtype: Vtype = Vtype.BINARY,
        variable_bounds: Union[List[Tuple[int, int]], List[Tuple[float, float]]]=[]
    ):
        """
        Add a vector of variable to the model. Multiple variables vector can be added but with different names.

        Notes
        -----
        If Vtype is set to ``Vtype.INTEGER`` or ``Vtype.CONTINUOUS``, variable_bounds need to be set.

        Parameters
        ----------
        name
            The name given to this variable vector.
        size
            The size of the vector.
        vtype
            Type of the variables inside the vector.
        variable_bounds
            Lower and upper bounds for the variable vector. A list of tuples (can be either integers or continuous)

        Raises
        ------
        MaximumVariableLimitError
            If the total size of variables exceed the limit.
        ValueError
            If the size of the vector is < 1

        Examples
        --------
        >>> from titanq import Model, Vtype
        >>> model.add_variable_vector('x', 3, Vtype.BINARY)
        >>> model.add_variable_vector('y', 3, Vtype.INTEGER, [[0, 5], [1, 6]])
        >>> model.add_variable_vector('z', 3, Vtype.CONTINUOUS, [[2.3, 4.6], [3.1, 5.3], [1.1, 4]])
        """
        # validation
        if not self._constraints.is_empty():
            raise ConstraintAlreadySetError("Cannot add additional variable once constraints have been defined")

        if self._objective is not None:
            raise ObjectiveAlreadySetError("Cannot add additional variable once objective have been defined")

        # create the Variable vector
        if vtype is Vtype.BINARY:
            if variable_bounds:
                raise ValueError("variable_bounds is not supported with Vtype.BINARY")
            variables = BinaryVariableVector(name, size)

        elif vtype is Vtype.BIPOLAR:
            if variable_bounds:
                raise ValueError("variable_bounds is not supported with Vtype.BIPOLAR")
            variables = BipolarVariableVector(name, size)

        elif vtype is Vtype.INTEGER:
            variables = IntegerVariableVector(name, size, variable_bounds)

        elif vtype is Vtype.CONTINUOUS:
            variables = ContinuousVariableVector(name, size, variable_bounds)

        else:
            raise NotImplementedError(f"Unsupported variable type: {vtype}")

        self._variables.add(variables)
        self._constraints.augment_size(size)

        log.debug(f"add variable name='{name}', type={str(vtype)}, size={size}.")


    def set_objective_matrices(self, weights: np.ndarray, bias: np.ndarray, target=Target.MINIMIZE):
        """
        Set the objective matrices for the model.

        Parameters
        ----------
        weights
            The quadratic objective matrix, **this matrix needs to be symmetrical**.
            A NumPy 2-D dense ndarray (must be float32).
        bias
            The linear constraint vector. A NumPy 1-D ndarray.
        target
            The target of this objective matrix.

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        ObjectiveAlreadySetError
            If an objective has already been set in this model.
        ValueError
            If the weights shape or the bias shape does not fit the variables in the model.
            If the weights or bias data type is not float32.

        Examples
        --------
        >>> from titanq import Model, Target
        >>> edges = {0:[4,5,6,7], 1:[4,5,6,7], 2:[4,5,6,7], 3:[4,5,6,7], 4:[0,1,2,3], 5:[0,1,2,3], 6:[0,1,2,3], 7:[0,1,2,3]}
        >>> size = len(edges)
        >>> weights = np.zeros((size, size), dtype=np.float32)
        >>> for root, connections in edges.items():
        >>>     for c in connections:
        >>>         weights[root][c] = 1
        >>> # construct the bias vector (Uniform weighting across all nodes)
        >>> bias = np.asarray([0]*size, dtype=np.float32)
        >>> model.set_objective_matrices(weights, bias, Target.MINIMIZE)
        """
        if self._variables.n_variables() == 0:
            raise MissingVariableError("Cannot set objective before adding a variable to the model.")

        if self._objective is not None:
            raise ObjectiveAlreadySetError("An objective has already have been set for this model.")

        log.debug(f"set objective matrix and bias vector.")

        self._objective = Objective(self._variables.total_variable_size(), weights, bias, target)


    def add_set_partitioning_constraints_matrix(self, constraint_mask: np.ndarray):
        """
        Adds set partitioning constraints in matrix format to the model.

        Parameters
        ----------
        constraint_mask
            A NumPy 2-D dense ndarray (must be binary).
            The constraint_mask matrix of shape (M, N) where M the number of constraints and N is the number of variables.

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MaximumConstraintLimitError
            The number of constraints exceed the limit.
        ConstraintSizeError
            If the constraint_mask shape does not fit the expected shape of this model.
        ValueError
            If the constraint_mask data type is not binary.

        Examples
        --------
        >>> constraint_mask = np.array([[1, 1, 1, 0, 1], [1, 1, 1, 1, 0]])
        >>> model.add_set_partitioning_constraints_matrix(constraint_mask)
        """
        if self._variables.n_variables() == 0:
            raise MissingVariableError("Cannot set constraints before adding a variable to the model.")

        if constraint_mask.ndim == 1:
            raise ValueError(
                "Cannot use add_set_partitioning_constraints_matrix() function with a vector, " \
                "please use add_set_partitioning_constraint() instead")

        if not is_ndarray_binary(constraint_mask):
            raise ValueError(f"Cannot add a constraint if the values are not in binary.")

        self._constraints.add_constraint(
            constraint_weights=constraint_mask,
            constraint_bounds=np.ones((constraint_mask.shape[0],2))
        )
        self._manifest.activate_set_partitioning_constraint()

    def add_set_partitioning_constraint(self, constraint_mask: np.ndarray):
        """
        Adds set partitioning constraint vector to the model.

        Parameters
        ----------
        constraint_mask
            A NumPy 1-D dense ndarray (must be binary).
            The constraint_mask vector of shape (N,) where N is the number of variables.

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MaximumConstraintLimitError
            The number of constraint exceed the limit.
        ConstraintSizeError
            If the constraint_mask shape does not fit the expected shape of this model.
        ValueError
            If the constraint_mask data type is not binary.

        Examples
        --------
        >>> constraint_mask = np.array([1, 1, 1, 0, 1])
        >>> model.add_set_partitioning_constraint(constraint_mask)
        """
        if constraint_mask.ndim > 1:
            raise ValueError(
                "Cannot use this add_set_partitioning_constraint() function with a matrix, " \
                "please use add_set_partitioning_constraints_matrix() instead")

        self.add_set_partitioning_constraints_matrix(reshape_to_2d(constraint_mask))


    def add_cardinality_constraints_matrix(self, constraint_mask: np.ndarray, cardinalities: np.ndarray):
        """
        Adds cardinality constraints in matrix format to the model.

        Parameters
        ----------
        constraint_mask
            A NumPy 2-D dense ndarray (must be binary).
            The constraint_mask matrix of shape (M, N) where M the number of constraints and N is the number of variables.
        cardinalities
            A NumPy 1-D ndarray (must be non-zero unsigned integer).
            The constraint_rhs vector of shape (M,) where M is the number of constraints.

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MaximumConstraintLimitError
            The number of constraint exceed the limit.
        ConstraintSizeError
            If the constraint_mask shape or the constraint_rhs shape does not fit the expected shape of this model.
        ValueError
            If the constraint_mask is not binary or cardinalities data type are not unsigned integers.

        Examples
        --------
        >>> constraint_mask = np.array([[1, 1, 1, 0, 1], [1, 1, 1, 1, 0]])
        >>> cardinalities = np.array([3, 2])
        >>> model.add_cardinality_constraints_matrix(constraint_mask, cardinalities)
        """
        if self._variables.n_variables() == 0:
            raise MissingVariableError("Cannot set constraints before adding a variable to the model.")

        if constraint_mask.ndim == 1:
            raise ValueError(
                "Cannot use add_cardinality_constraints_matrix() function with a vector, " \
                "please use add_cardinality_constraint() instead")

        if cardinalities.ndim != 1:
            raise ValueError(f"Cannot set constraints if cardinalities is not a NumPy 1-D dense ndarray")

        if not np.issubdtype(cardinalities.dtype, np.integer):
            raise ValueError("Found cardinalities data types not integer")

        if not np.all(cardinalities > 0):
            raise ValueError("Found cardinalities data types not unsigned integer")

        if not is_ndarray_binary(constraint_mask):
            raise ValueError(f"Cannot add a constraint if the values are not in binary.")

        if cardinalities.shape[0] != constraint_mask.shape[0]:
            raise ValueError(
                f"Cannot set constraints if cardinalities shape is not the same as the expected shape of this model." \
                f" Got cardinalities shape: {cardinalities.shape[1]}, constraint mask shape: {constraint_mask.size()}.")

        validate_cardinalities(constraint_mask, cardinalities)

        self._constraints.add_constraint(
            constraint_weights=constraint_mask,
            constraint_bounds=np.repeat(cardinalities, 2).reshape(-1, 2)
        )
        self._manifest.activate_cardinality_constraint()


    def add_cardinality_constraint(self, constraint_mask: np.ndarray, cardinality: int):
        """
        Adds cardinality constraint vector to the model.

        Parameters
        ----------
        constraint_mask
            A NumPy 1-D dense ndarray (must be binary).
            The constraint_mask vector of shape (N,) where N is the number of variables.
        cardinality
            The constraint_rhs cardinality.
            This value has to be a non-zero unsigned integer.

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MaximumConstraintLimitError
            If the number of constraints exceed the limit.
        ConstraintSizeError
            If the constraint_mask shape or the constraint_rhs shape does not fit
            the expected shape of this model.
        ValueError
            If the constraint_mask is not in binary or the cardinality is not an unsigned integer.

        Examples
        --------
        >>> constraint_mask = np.array([1, 1, 1, 0, 1])
        >>> cardinality = 3
        >>> model.add_cardinality_constraint(constraint_mask, cardinality)
        """
        if constraint_mask.ndim > 1:
            raise ValueError(
                "Cannot use add_cardinality_constraint() function with a matrix, " \
                "please use add_cardinality_constraints_matrix() instead")

        self.add_cardinality_constraints_matrix(reshape_to_2d(constraint_mask), np.full((1,), cardinality))


    def add_equality_constraints_matrix(self, constraint_mask: np.ndarray, limit: np.ndarray) -> None:
        """
        Adds an equality constraint matrix to the model.

        Parameters
        ----------
        constraint_mask
            A NumPy 2-D dense ndarray (float32).
            The constraint_mask vector of shape (M, N) where M the number of constraints and N is the number of variables.
        limit
            A NumPy 1-D array (float32).
            The limit vector of shape (M,) where M is the number of constraints.

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MaximumConstraintLimitError
            The number of constraint exceed the limit.
        ValueError
            If the constraint_mask shape does not fit the expected shape of this model.
            If the constraint_mask or limit contains irregular format ('NaN' or 'inf').

        Examples
        --------
        >>> constraint_mask = np.array([[-3.51, 0, 0, 0], [10, 0, 0, 0]], dtype=np.float32)
        >>> limit = np.array([2, 10], dtype=np.float32)
        >>> model.add_equality_constraints_matrix(constraint_mask, limit)
        """
        if self._variables.n_variables() == 0:
            raise MissingVariableError("Cannot set constraints before adding a variable to the model.")

        if constraint_mask.dtype != np.float32 or limit.dtype != np.float32:
            raise ValueError(f"Input parameters must be float32, got Constraint mask: {constraint_mask.dtype}, Limit: {limit.dtype}")

        if constraint_mask.ndim == 1:
            raise ValueError(
                "Cannot use add_equality_constraint_matrix() function with a vector, " \
                "please use add_equality_constraint() instead")

        if ndarray_contains_nan_inf(constraint_mask):
            raise ValueError("Constraint mask contains NaN  or inf values")

        if limit.ndim != 1:
            raise ValueError("Limit must be a NumPy 1-D array")

        if ndarray_contains_nan_inf(limit):
            raise ValueError("Limit contains NaN or inf values")

        if self._variables.has_variable_of_type(Vtype.BIPOLAR):
            raise ValueError("Cannot add equality constraint with variable vector set to 'bipolar'")

        self._constraints.add_constraint(
            constraint_weights=constraint_mask,
            constraint_bounds=np.repeat(limit, 2).reshape(-1, 2)
        )
        self._manifest.activate_equality_constraint()

    def add_equality_constraint(self, constraint_mask: np.ndarray, limit: np.float32) -> None:
        """
        Adds an equality constraint vector to the model.

        Parameters
        ----------
        constraint_mask
            A NumPy 1-D dense ndarray (float32).
            The constraint_mask vector of shape (N,) where N is the number of variables.
        limit
            Limit value to the constraint mask.

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MaximumConstraintLimitError
            The number of constraint exceed the limit.
        ValueError
            If the constraint_mask shape does not fit the expected shape of this model.
            If the constraint_mask or limit contains irregular format ('NaN' or 'inf').

        Examples
        --------
        >>> constraint_mask = np.array([1.05, -1.1], dtype=np.float32)
        >>> limit = -3.45
        >>> model.add_equality_constraint(constraint_mask, limit)
        """
        if constraint_mask.ndim > 1:
            raise ValueError(
                "Cannot use add_equality_constraint() function with a matrix, " \
                "please use add_equality_constraint_matrix() instead")

        self.add_equality_constraints_matrix(reshape_to_2d(constraint_mask), np.full((1,), limit, dtype=np.float32))


    def add_inequality_constraints_matrix(self, constraint_mask: np.ndarray, constraint_bounds: np.ndarray):
        """
        Adds inequality constraint matrix to the model.

        Parameters
        ----------
        constraint_mask
            A NumPy 2-D dense ndarray (float32).
            The constraint_mask vector of shape (M, N) where N is the number of variables.
        constraint_bounds
            A NumPy 2-D ndarray (float32).
            Vector of shape (M, 2) where M is the number of constraints.

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MaximumConstraintLimitError
            The number of constraint exceed the limit.
        ValueError
            If the constraint_mask shape does not fit the expected shape of this model.
            If the constraint_mask contains irregular format ('NaN' or 'inf').
            If the lowerbound is equal or higher than its given upperbound.

        Examples
        --------
        >>> constraint_mask = np.array([[-3.51, 0], [10, 0]], dtype=np.float32)
        >>> constraint_bounds = np.array([[8, 9], [np.nan, 100_000]], dtype=np.float32)
        >>> model.add_inequality_constraints_matrix(constraint_mask, constraint_bounds)
        """
        if self._variables.n_variables() == 0:
            raise MissingVariableError("Cannot set constraints before adding a variable to the model.")

        if constraint_mask.dtype != np.float32 or constraint_bounds.dtype != np.float32:
            raise ValueError(f"Input parameters must be float32, got Constraint mask: {constraint_mask.dtype}, Limit: {constraint_bounds.dtype}")

        if constraint_mask.ndim == 1:
            raise ValueError(
                "Cannot use add_inequality_constraint_matrix() function with a vector, " \
                "please use add_inequality_constraint() instead")

        if ndarray_contains_nan_inf(constraint_mask):
            raise ValueError("Constraint mask contains NaN  or inf values.")

        if is_upperbound_bigger_equal_lowerbound(constraint_bounds):
            raise ValueError("Constraint bounds contains lowerbounds equal or higher than their upperbound.")

        self._constraints.add_constraint(
            constraint_weights=constraint_mask,
            constraint_bounds=constraint_bounds
        )
        self._manifest.activate_inequality_constraint()


    def add_inequality_constraint(self, constraint_mask: np.ndarray, constraint_bounds: np.ndarray):
        """
        Adds inequality constraint vector to the model. At least one bound must be set.

        Parameters
        ----------
        constraint_mask
            A NumPy 1-D dense ndarray (float32).
            The constraint_mask vector of shape (N,) where N is the number of variables.
        constraint_bounds
            A NumPy 1-D ndarray (float32).
            Vector of shape (2,)

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MaximumConstraintLimitError
            The number of constraint exceed the limit.
        ValueError
            If the constraint_mask shape does not fit the expected shape of this model.
            If the constraint_mask contains irregular format ('NaN' or 'inf').
            If the lowerbound is equal or higher than the upperbound.

        Examples
        --------
        >>> constraint_mask = np.array([1.05, -1.1], dtype=np.float32)
        >>> constraint_bounds = np.array([1.0, np.nan], dtype=np.float32)
        >>> model.add_inequality_constraint(constraint_mask, constraint_bounds)
        """
        if constraint_mask.ndim > 1:
            raise ValueError(
                "Cannot use add_inequality_constraint() function with a matrix, " \
                "please use add_inequality_constraint_matrix() instead")

        self.add_inequality_constraints_matrix(reshape_to_2d(constraint_mask), reshape_to_2d(constraint_bounds))


    def optimize(
        self,
        *,
        beta: List[float] = [0.1],
        coupling_mult: float = 0.5,
        timeout_in_secs: float = 10.0,
        num_chains: int = 8,
        num_engines: int = 1,
        normalized: bool = False
    ) -> OptimizeResponse:
        """
        Optimize this model.

        Notes
        -----
        All of the files used during this computation will be cleaned at the end.
        For more information on how to tunes those parameters, visit the API doc at `TitanQ API <https://docs.titanq.infinityq.io/>`_.

        Parameters
        ----------
        beta
            ``beta`` hyper parameter used by the solver.
        coupling_mult
            ``coupling_mult`` hyper parameter used by the solver.
        timeout_in_secs
            Maximum time (in seconds) the solver can take to solve this problem.
        num_chains
            ``num_chains`` hyper parameter used by the solver.
        num_engines
            ``num_engines`` parameter used by the solver.
        normalized
            **DEPRECATED since version 0.9.1**. ``normalized`` should only be used when the model is only using
            binary or bipolar variables and does not contain any constraints.

        Returns
        -------
        OptimizeResponse
            Optimized response data object

        Raises
        ------
        MissingVariableError
            If no variable have been added to the model.
        MissingObjectiveError
            If no objective matrices have been added to the model.

        Examples
        --------
        basic solve
            >>> response = model.optimize(timeout_in_secs=60)
        multiple engine
            >>> response = model.optimize(timeout_in_secs=60, num_engines=2)
        custom values
            >>> response = model.optimize(beta=[0.1], coupling_mult=0.75, num_chains=8)
        print values
            >>> print("-" * 15, "+", "-" * 26, sep="")
            >>> print("Ising energy   | Result vector")
            >>> print("-" * 15, "+", "-" * 26, sep="")
            >>> for ising_energy, result_vector in response.result_items():
            >>>     print(f"{ising_energy: <14f} | {result_vector}")
        """
        if self._variables.n_variables() == 0:
            raise MissingVariableError("Cannot optimize before adding a variable to the model.")

        if self._objective is None:
            raise MissingObjectiveError("Cannot optimize before adding an objective to the model.")

        result, metrics = self._solve(beta, coupling_mult, timeout_in_secs, num_chains, num_engines, normalized)

        return OptimizeResponse(self._variables, result, metrics)

    def _solve(self,
            beta: List[float],
            coupling_mult: float,
            timeout_in_secs: float,
            num_chains: int,
            num_engines: int,
            normalized: bool
        ) -> Tuple[np.ndarray, Dict[str, Any]]:
        """
        issue a solve request and wait for it to complete.

        Parameters
        ----------
        beta
            beta hyper parameter used by the backend solver.
        coupling_mult
            coupling_mult hyper parameter used by the backend solver.
        timeout_in_secs
            Maximum time (in seconds) the backend solver can take to resolve this problem.
        num_chains
            num_chains hyper parameter used by the backend solver.
        num_engines
            num_engines parameter used by the backend solver.
        normalized

        Returns
        -------
        The result numpy array and the metric json object.
        """
        with self._storage_client.temp_files_manager(
            self._objective.bias(),
            self._objective.weights(),
            self._constraints.bounds() if self._constraints else None,
            self._constraints.weights() if self._constraints else None,
            # variables bounds is only sent if variable is set to binary
            self._variables.variable_bounds() if not self._variables.has_variable_of_type(Vtype.BIPOLAR) else None
        ) as temp_files:

            # api request
            request = api_model.SolveRequest(
                input=temp_files.input(),
                output=temp_files.output(),
                parameters=api_model.Parameters(
                    beta=beta,
                    coupling_mult=coupling_mult,
                    num_chains=num_chains,
                    num_engines=num_engines,
                    normalized=normalized,
                    timeout_in_secs=timeout_in_secs,
                    variable_types=self._variables.get_variable_types_str(self._manifest.has_equality_constraint(), self._manifest.has_inequality_constraint())
                )
            )

            # adding manifest flags to the request
            request.input.manifest = api_model.Manifest(
                has_set_partitioning_constraint=self._manifest.has_set_partitioning_constraint(),
                has_cardinality_constraint=self._manifest.has_cardinality_constraint(),
                has_equality_constraint=self._manifest.has_equality_constraint(),
                has_inequality_constraint=self._manifest.has_inequality_constraint()
            )

            solve_response = self._titanq_client.solve(request)

            # wait for result to be uploaded by the solver and download it
            archive_file_content = temp_files.download_result()
            with zipfile.ZipFile(io.BytesIO(archive_file_content), 'r') as zip_file:
                try:
                    metrics_content = zip_file.read("metrics.json")
                    result_content = zip_file.read("result.npy")
                except KeyError as ex:
                    try:
                        error_content = zip_file.read("error.json")
                        raise OptimizeError(json.loads(error_content)["error"]) from ex
                    except KeyError as e:
                        raise OptimizeError(
                            "Unexpected error in the solver, please contact InfinityQ support for more help" \
                            f" and provide the following computation id {solve_response.computation_id}") from e

        log.debug("Optimization completed")
        return np.load(io.BytesIO(result_content)), json.loads(metrics_content)



import logging
import requests

from pydantic import BaseModel, ValidationError
from typing import Any, Dict, Type
from urllib.parse import urljoin

from .model import CreditsResponse, SolveRequest, SolveResponse, TempStorageResponse
from .._model.errors import ServerError

log = logging.getLogger("TitanQ")

_QUEUED_STATUS = "Queued"
_TITANQ_API_VERSION = "v1"

class Client:
    """
    TitanQ api client is a simple wrapper around TitanQ api to help interact with the
    service without the need to deal with http request
    """
    def __init__(self, api_key: str, base_server_url: str) -> None:
        self._server_url = base_server_url
        self._api_key = api_key


    def temp_storage(self) -> TempStorageResponse:
        """
        Query temporary storage url's

        :return: The temporary storage response

        :raises requests.exceptions.HTTPError: If an unexpected Error occur during request.
        """
        return self._do_http_request(f"{_TITANQ_API_VERSION}/temp_storage", response_type=TempStorageResponse)

    def credits(self) -> CreditsResponse:
        """
        Query Amount of credits remaining

        :return: The credit response.

        :raises requests.exceptions.HTTPError: If an unexpected Error occur during request.
        """
        return self._do_http_request(f"{_TITANQ_API_VERSION}/credits", response_type=CreditsResponse)

    def solve(self, request: SolveRequest) -> SolveResponse:
        """
        Issue a new solve request to the backend

        :param request: The solve request to issue to the solver

        :return: The response to the solve request (Not the response of the computation)

        :raises ServerError: If an unexpected Error occur during a solver request
        """
        log.debug(f"Issuing solve request to TitanQ server ({self._server_url}): {request}")
        response = self._do_http_request(f"{_TITANQ_API_VERSION}/solve", body=request, method='POST', response_type=SolveResponse)

        # something went wrong and the computation was not queued
        if response.status != _QUEUED_STATUS:
            log.error("An error occured while issuing a solver request to the TitanQ server")
            raise ServerError(response.message)

        log.debug(f"Solve request response: {response}")
        return response


    def _do_http_request(
            self,
            path: str,
            *,
            headers: Dict[str, str] = {},
            body: BaseModel = None,
            method='GET',
            response_type: Type
        ) -> Any:
        """
        Execute the actual http request to the TitanQ api while adding all defaults params

        :param headers: non-default header to the request.
        :param body: Body of the request.
        :param method: Which http method to use while performing the request.
        :param response_type: The object class that the json response will be cast to.

        :raise HTTPError: If the response cannot be created from the response type passed.

        :return: The response object created from the json response of the http request.
        """
        headers['authorization'] = self._api_key
        url = urljoin(self._server_url, path)

        method = method.upper()
        if method=='GET':
            response = requests.get(url, headers=headers)
        elif method=='POST':
            response = requests.post(url, headers=headers, data=body.model_dump_json())
        else:
            raise NotImplementedError(f"http method: {method}")

        try:
            # create the response object from the response body
            return response_type.model_validate_json(response.content)
        except ValidationError:
            response.raise_for_status()
            raise


import numpy as np

from abc import ABC, abstractmethod
from io import BytesIO
from typing import Optional, Type, Union


from .._client.model import S3Input, S3Output, UrlInput, UrlOutput


class StorageClient(ABC):
    def __init__(self):
        pass

    def temp_files_manager(self,
        bias: np.ndarray,
        weights: np.ndarray,
        constraint_bounds: Optional[np.ndarray],
        constraint_weights: Optional[np.ndarray],
        variable_bounds: Optional[np.ndarray]
    ) -> Type["TempFileManager"]:
        """'
        Creates and returns a context manager that will be handling the temporary file management

        :param bias: bias numpy array
        :param weights: weights numpy array
        """
        return TempFileManager(self, bias, weights, constraint_bounds, constraint_weights, variable_bounds)

    @abstractmethod
    def _input(self) -> Union[S3Input, UrlInput]:
        """
        Returns the api model for the input of the solve request

        :return: either the s3 or the url input
        """

    @abstractmethod
    def _output(self) -> Union[S3Output, UrlOutput]:
        """
        Returns the api model for the output of the solve request

        :return: either the s3 or the url output
        """

    @abstractmethod
    def _upload_arrays(
        self,
        bias: bytes,
        weights: bytes,
        constraint_bounds: Optional[bytes],
        constraint_weights: Optional[bytes],
        variable_bounds: Optional[bytes]
        ):
        """Uploads .npy arrays to the storage client"""

    @abstractmethod
    def _wait_for_result_to_be_uploaded_and_download(self) -> bytes:
        """
        Wait until a file is uploaded on the storage client and download it

        :return: content of the result file in bytes
        """

    @abstractmethod
    def _delete_remote_object(self) -> None:
        """
        Deletes a remote object on the storage client
        """

class TempFileManager():
    """
    Context manager handling the upload and download to a temporary storage
    """
    def __init__(
        self,
        storage_client: StorageClient,
        bias: np.ndarray,
        weights: np.ndarray,
        constraint_bounds: Optional[np.ndarray],
        constraint_weights: Optional[np.ndarray],
        variable_bounds: Optional[np.ndarray]
    ) -> None:
        """
        :param storage_client: storage client that will handle uploads, downloads and generating api models
        :param bias: bias numpy array
        :param weights: weights numpy array
        """
        super().__init__()
        self._storage_client = storage_client
        self._bias = bias
        self._weights = weights
        self._constraint_bounds = constraint_bounds
        self._constraint_weights = constraint_weights
        self._variable_bounds = variable_bounds

    def __enter__(self) -> Type["TempFileManager"]:
        self._storage_client._upload_arrays(
            self._to_bytes(self._bias),
            self._to_bytes(self._weights),
            self._to_bytes(self._constraint_bounds),
            self._to_bytes(self._constraint_weights),
            self._to_bytes(self._variable_bounds)
        )
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self._storage_client._delete_remote_object()

    def input(self) -> Union[S3Input, UrlInput]:
        return self._storage_client._input()

    def output(self) -> Union[S3Output, UrlOutput]:
        return self._storage_client._output()

    def download_result(self):
        return self._storage_client._wait_for_result_to_be_uploaded_and_download()

    def _to_bytes(self, array: Optional[np.ndarray]) -> Optional[bytes]:
        """
        :return: numpy array as bytes
        """
        if array is None:
            return None

        buffer = BytesIO()
        np.save(buffer, array)
        return buffer.getvalue()


import numpy as np
import pytest

from titanq._model.variable import BinaryVariableVector, BipolarVariableVector, ContinuousVariableVector, IntegerVariableVector, Vtype


@pytest.mark.parametrize("variable_vector, expected",  [
    (BinaryVariableVector('x', 3), Vtype.BINARY),
    (BipolarVariableVector('x', 3), Vtype.BIPOLAR),
    (IntegerVariableVector('x', 3, [(1, 2)] * 3), Vtype.INTEGER),
    (ContinuousVariableVector('x', 3, [(1, 2)] * 3), Vtype.CONTINUOUS),
])
def test_variable_types(variable_vector, expected):
    assert variable_vector.vtype() == expected


@pytest.mark.parametrize("variable_vector, expected",  [
    (BinaryVariableVector('x', 3), "bbb"),
    (IntegerVariableVector('x', 4, [(1, 2)] * 4), "iiii"),
    (ContinuousVariableVector('x', 8, [(2, 6)] * 8), "cccccccc"),
])
def test_variable_types_as_list(variable_vector, expected):
    assert variable_vector.variable_types_as_list() == expected


def test_binary_variable_bounds():
    variable_vector = BinaryVariableVector("x", 3)
    assert (np.array([[0, 1], [0, 1], [0, 1]]) == variable_vector.variable_bounds()).all()

